"""
Created on Tue Dec 21 16:54:50 2021

@author: stardist devs
"""

# -*- coding: utf-8 -*-

from collections import OrderedDict
from warnings import warn
from csbdeep.utils import _raise
from csbdeep.utils.six import Path

from csbdeep.utils.tf import keras_import

get_file = keras_import("utils", "get_file")


_MODELS = {}
_ALIASES = {}


def clear_models_and_aliases(*cls):
    if len(cls) == 0:
        _MODELS.clear()
        _ALIASES.clear()
    else:
        for c in cls:
            if c in _MODELS:
                del _MODELS[c]
            if c in _ALIASES:
                del _ALIASES[c]


def register_model(cls, key, url, hash):
    # key must be a valid file/folder name in the file system
    models = _MODELS.setdefault(cls, OrderedDict())
    key not in models or warn(
        "re-registering model '{}' (was already registered for '{}')".format(
            key, cls.__name__
        )
    )
    models[key] = dict(url=url, hash=hash)


def register_aliases(cls, key, *names):
    # aliases can be arbitrary strings
    if len(names) == 0:
        return
    models = _MODELS.get(cls, {})
    key in models or _raise(
        ValueError(f"model '{key}' is not registered for '{cls.__name__}'")
    )
    aliases = _ALIASES.setdefault(cls, OrderedDict())
    for name in names:
        aliases.get(name, key) == key or warn(
            "alias '{}' was previously registered with model '{}' for '{}'".format(
                name, aliases[name], cls.__name__
            )
        )
        aliases[name] = key


def get_registered_models(cls, return_aliases=True, verbose=False):
    models = _MODELS.get(cls, {})
    aliases = _ALIASES.get(cls, {})
    model_keys = tuple(models.keys())
    model_aliases = {
        key: tuple(name for name in aliases if aliases[name] == key) for key in models
    }
    if verbose:
        # this code is very messy and should be refactored...
        _n = len(models)
        _str_model = "model" if _n == 1 else "models"
        _str_is_are = "is" if _n == 1 else "are"
        _str_colon = ":" if _n > 0 else ""
        print(
            "There {is_are} {n} registered {model_s} for '{clazz}'{c}".format(
                n=_n,
                clazz=cls.__name__,
                is_are=_str_is_are,
                model_s=_str_model,
                c=_str_colon,
            )
        )
        if _n > 0:
            print()
            _maxkeylen = 2 + max(len(key) for key in models)
            print("Name{s}Alias(es)".format(s=" " * (_maxkeylen - 4 + 3)))
            print("────{s}─────────".format(s=" " * (_maxkeylen - 4 + 3)))
            for key in models:
                _aliases = "   "
                _m = len(model_aliases[key])
                if _m > 0:
                    _aliases += "'%s'" % "', '".join(model_aliases[key])
                else:
                    _aliases += "None"
                _key = ("{s:%d}" % _maxkeylen).format(s="'%s'" % key)
                print(f"{_key}{_aliases}")
    return (model_keys, model_aliases) if return_aliases else model_keys


def get_model_details(cls, key_or_alias, verbose=False):
    models = _MODELS.get(cls, {})
    if key_or_alias in models:
        key = key_or_alias
        alias = None
    else:
        aliases = _ALIASES.get(cls, {})
        alias = key_or_alias
        alias in aliases or _raise(
            ValueError(f"'{alias}' is neither a key or alias for '{cls.__name__}'")
        )
        key = aliases[alias]
    if verbose:
        print(
            "Found model '{model}'{alias_str} for '{clazz}'.".format(
                model=key,
                clazz=cls.__name__,
                alias_str=("" if alias is None else " with alias '%s'" % alias),
            )
        )
    return key, alias, models[key]


def get_model_folder(cls, key_or_alias):
    key, alias, m = get_model_details(cls, key_or_alias)
    target = str(Path("models") / cls.__name__ / key)
    path = Path(
        get_file(
            fname=key + ".zip",
            origin=m["url"],
            file_hash=m["hash"],
            cache_subdir=target,
            extract=True,
        )
    )
    assert path.exists() and path.parent.exists()
    return path.parent


def get_model_instance(cls, key_or_alias):
    path = get_model_folder(cls, key_or_alias)
    model = cls(config=None, name=path.stem, basedir=path.parent)
    model.basedir = None  # make read-only
    return model



import os
from pathlib import Path
from .cellposeutils3D import (
    save_json,
)
from .TrainTiledLoader import TrainTiled
from torch.utils.data import DataLoader
from .unet3d import ResidualUNet3D, UNet3D
import torch
import torch.nn.functional as F
from torch import optim
from collections import OrderedDict
from lightning import Trainer, LightningModule
from lightning.pytorch.loggers import CSVLogger
from lightning.pytorch.callbacks import ModelCheckpoint
from .cellposeutils3D import prepare_images, prepare_masks, create_csv

torch.set_float32_matmul_precision("medium")


class CellPose3DPredict(LightningModule):
    def __init__(self, model, hparams):
        super().__init__()
        self.save_hyperparameters(hparams)
        self.model = model

    def predict_step(self, batch, batch_idx):

        self.tile, self.coord = batch

        self.pred = self.model(self.tile.float())

        return self.pred, self.coord


class CellPose3DModel(LightningModule):
    def __init__(
        self,
        hparams,
    ):
        super().__init__()

        self.save_hyperparameters(hparams)
        self.network = UNet3D(
            in_channels=hparams["in_channels"],
            out_channels=hparams["out_channels"],
            f_maps=hparams["feat_channels"],
            num_levels=hparams["num_levels"],
        )

        self.in_channels = hparams["in_channels"]
        self.out_channels = hparams["out_channels"]
        self.patch_size = hparams["patch_size"]
        self.background_weight = hparams["background_weight"]
        self.flow_weight = hparams["flow_weight"]
        self.learning_rate = hparams["learning_rate"]

    def load_pretrained(self, pretrained_file, strict=True, verbose=True):
        if isinstance(pretrained_file, (list, tuple)):
            pretrained_file = pretrained_file[0]

        # Load the state dict
        state_dict = torch.load(pretrained_file)["state_dict"]

        # Make sure to have a weight dict
        if not isinstance(state_dict, dict):
            state_dict = dict(state_dict)

        # Get parameter dict of current model
        param_dict = dict(self.network.named_parameters())

        layers = []
        for layer in param_dict:
            if strict and not "network." + layer in state_dict:
                if verbose:
                    print(f'Could not find weights for layer "{layer}"')
                continue
            try:
                param_dict[layer].data.copy_(
                    state_dict["network." + layer].data
                )
                layers.append(layer)
            except (RuntimeError, KeyError) as e:
                print(f"Error at layer {layer}:\n{e}")

        self.network.load_state_dict(param_dict)

        if verbose:
            print(f"Loaded weights for the following layers:\n{layers}")

    def background_loss(self, y_hat, y):
        return F.l1_loss(y_hat, y)

    def flow_loss(self, y_hat, y, mask):
        loss = F.mse_loss(y_hat, y, reduction="none")
        weight = torch.clamp(mask, min=0.01, max=1.0)
        loss = torch.mul(loss, weight)
        loss = torch.sum(loss)
        loss = torch.div(loss, torch.clamp(torch.sum(weight), 1, mask.numel()))
        return loss

    def forward(self, z):
        return self.network(z)

    def training_step(self, batch, batch_idx):
        # Get image ans mask of current batch
        self.last_imgs, self.last_masks = batch["image"], batch["mask"]

        # generate images
        self.predictions = self.forward(self.last_imgs)

        # get the losses
        loss_bg = self.background_loss(
            self.predictions[:, 0, ...], self.last_masks[:, 0, ...]
        )

        loss_flowx = self.flow_loss(
            self.predictions[:, 1, ...],
            self.last_masks[:, 1, ...],
            self.last_masks[:, 0, ...],
        )
        loss_flowy = self.flow_loss(
            self.predictions[:, 2, ...],
            self.last_masks[:, 2, ...],
            self.last_masks[:, 0, ...],
        )
        loss_flowz = self.flow_loss(
            self.predictions[:, 3, ...],
            self.last_masks[:, 3, ...],
            self.last_masks[:, 0, ...],
        )
        loss_flow = (loss_flowx + loss_flowy + loss_flowz) / 3

        loss = self.background_weight * loss_bg + self.flow_weight * loss_flow

        return loss

    def test_step(self, batch, batch_idx):

        self._shared_eval(batch, batch_idx, "test")

    def _shared_eval(self, batch, batch_idx, prefix):

        x, y = batch["image"], batch["mask"]
        y_hat = self.forward(x)
        loss = F.mse_loss(y_hat, y)
        self.log(
            f"{prefix}_loss", loss, on_step=True, on_epoch=True, sync_dist=True
        )

    def validation_step(self, batch, batch_idx):

        self._shared_eval(batch, batch_idx, "validation")

    def configure_optimizers(self):
        optimizer = optim.Adam(
            self.network.parameters(), lr=self.learning_rate
        )

        schedular = optim.lr_scheduler.ReduceLROnPlateau(
            optimizer=optimizer, factor=0.5
        )
        optimizer_scheduler = OrderedDict(
            {
                "optimizer": optimizer,
                "lr_scheduler": {
                    "scheduler": schedular,
                    "monitor": "validation_loss",
                    "frequency": 1,
                },
            }
        )
        return optimizer_scheduler


class CellPoseRes3DModel(LightningModule):
    def __init__(
        self,
        hparams,
    ):
        super().__init__()

        self.save_hyperparameters(hparams)
        self.network = ResidualUNet3D(
            in_channels=hparams["in_channels"],
            out_channels=hparams["out_channels"],
            f_maps=hparams["feat_channels"],
            num_levels=hparams["num_levels"],
        )

        self.in_channels = hparams["in_channels"]
        self.out_channels = hparams["out_channels"]
        self.patch_size = hparams["patch_size"]
        self.background_weight = hparams["background_weight"]
        self.flow_weight = hparams["flow_weight"]
        self.learning_rate = hparams["learning_rate"]

    def load_pretrained(self, pretrained_file, strict=True, verbose=True):
        if isinstance(pretrained_file, (list, tuple)):
            pretrained_file = pretrained_file[0]

        # Load the state dict
        state_dict = torch.load(pretrained_file)["state_dict"]

        # Make sure to have a weight dict
        if not isinstance(state_dict, dict):
            state_dict = dict(state_dict)

        # Get parameter dict of current model
        param_dict = dict(self.network.named_parameters())

        layers = []
        for layer in param_dict:
            if strict and not "network." + layer in state_dict:
                if verbose:
                    print(f'Could not find weights for layer "{layer}"')
                continue
            try:
                param_dict[layer].data.copy_(
                    state_dict["network." + layer].data
                )
                layers.append(layer)
            except (RuntimeError, KeyError) as e:
                print(f"Error at layer {layer}:\n{e}")

        self.network.load_state_dict(param_dict)

        if verbose:
            print(f"Loaded weights for the following layers:\n{layers}")

    def background_loss(self, y_hat, y):
        return F.l1_loss(y_hat, y)

    def flow_loss(self, y_hat, y, mask):
        loss = F.mse_loss(y_hat, y, reduction="none")
        weight = torch.clamp(mask, min=0.01, max=1.0)
        loss = torch.mul(loss, weight)
        loss = torch.sum(loss)
        loss = torch.div(loss, torch.clamp(torch.sum(weight), 1, mask.numel()))
        return loss

    def forward(self, z):
        return self.network(z)

    def training_step(self, batch, batch_idx):
        # Get image ans mask of current batch
        self.last_imgs, self.last_masks = batch["image"], batch["mask"]

        # generate images
        self.predictions = self.forward(self.last_imgs)

        # get the losses
        loss_bg = self.background_loss(
            self.predictions[:, 0, ...], self.last_masks[:, 0, ...]
        )

        loss_flowx = self.flow_loss(
            self.predictions[:, 1, ...],
            self.last_masks[:, 1, ...],
            self.last_masks[:, 0, ...],
        )
        loss_flowy = self.flow_loss(
            self.predictions[:, 2, ...],
            self.last_masks[:, 2, ...],
            self.last_masks[:, 0, ...],
        )
        loss_flowz = self.flow_loss(
            self.predictions[:, 3, ...],
            self.last_masks[:, 3, ...],
            self.last_masks[:, 0, ...],
        )
        loss_flow = (loss_flowx + loss_flowy + loss_flowz) / 3

        loss = self.background_weight * loss_bg + self.flow_weight * loss_flow

        return loss

    def test_step(self, batch, batch_idx):

        self._shared_eval(batch, batch_idx, "test")

    def _shared_eval(self, batch, batch_idx, prefix):

        x, y = batch["image"], batch["mask"]
        y_hat = self.forward(x)
        loss = F.mse_loss(y_hat, y)
        self.log(
            f"{prefix}_loss", loss, on_step=True, on_epoch=True, sync_dist=True
        )

    def validation_step(self, batch, batch_idx):

        self._shared_eval(batch, batch_idx, "validation")

    def configure_optimizers(self):
        optimizer = optim.Adam(
            self.network.parameters(), lr=self.learning_rate
        )

        schedular = optim.lr_scheduler.ReduceLROnPlateau(
            optimizer=optimizer, factor=0.5
        )
        optimizer_scheduler = OrderedDict(
            {
                "optimizer": optimizer,
                "lr_scheduler": {
                    "scheduler": schedular,
                    "monitor": "validation_loss",
                    "frequency": 1,
                },
            }
        )
        return optimizer_scheduler


class CellPose3DTrain:
    def __init__(
        self,
        base_dir,
        model_dir,
        model_name,
        patch_size=(8, 256, 256),
        epochs=100,
        in_channels=1,
        out_channels=4,
        feat_channels=[32, 64, 128, 256],
        num_levels=4,
        samples_per_epoch=-1,
        batch_size=16,
        learning_rate=0.001,
        background_weight=1,
        flow_weight=1,
        out_activation="tanh",
        raw_dir="raw/",
        real_mask_dir="real_mask/",
        identifier="*.tif",
        save_train="_train.csv",
        save_test="_test.csv",
        save_val="_val.csv",
        axis_norm=(0, 1, 2),
        variance_size=(5, 5, 5),
        fg_footprint_size=5,
        bg_label=0,
        channel=0,
        corrupt_prob=0,
        num_gpu=1,
        zoom_factor=(1, 1, 1),
        ckpt_path=None,
        network_type="resunet",
    ):

        self.base_dir = base_dir
        self.model_dir = model_dir
        self.model_name = model_name
        self.patch_size = patch_size
        self.epochs = epochs
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.feat_channels = feat_channels
        self.num_levels = num_levels
        self.samples_per_epoch = samples_per_epoch
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.background_weight = background_weight
        self.flow_weight = flow_weight
        self.out_activation = out_activation
        self.raw_dir = os.path.join(base_dir, raw_dir)
        self.real_mask_dir = os.path.join(base_dir, real_mask_dir)
        self.identifier = identifier
        self.save_train = save_train
        self.save_test = save_test
        self.save_val = save_val
        self.axis_norm = axis_norm
        self.variance_size = variance_size
        self.fg_footprint_size = fg_footprint_size
        self.channel = channel
        self.bg_label = bg_label
        self.corrupt_prob = corrupt_prob
        self.zoom_factor = zoom_factor
        self.num_gpu = num_gpu
        self.save_raw_h5_name = "raw_h5/"
        self.save_real_mask_h5_name = "real_mask_h5/"
        self.ckpt_path = ckpt_path
        self.network_type = network_type
        torch.cuda.empty_cache()

    def _create_training_h5(self):

        self.save_raw_h5 = os.path.join(self.base_dir, self.save_raw_h5_name)
        self.save_real_mask_h5 = os.path.join(
            self.base_dir, self.save_real_mask_h5_name
        )
        prepare_images(
            data_path=self.raw_dir,
            save_path=self.save_raw_h5,
            identifier=self.identifier,
            axis_norm=self.axis_norm,
            get_distance=False,
            get_illumination=False,
            get_variance=False,
            variance_size=self.variance_size,
            fg_footprint_size=self.fg_footprint_size,
            channel=self.channel,
        )

        prepare_masks(
            data_path=self.real_mask_dir,
            save_path=self.save_real_mask_h5,
            identifier=self.identifier,
            bg_label=self.bg_label,
            get_flows=True,
            get_boundary=False,
            get_seeds=False,
            get_distance=True,
            corrupt_prob=self.corrupt_prob,
            zoom_factor=self.zoom_factor,
        )
        data_list = []
        for imagename in os.listdir(self.save_raw_h5):
            data_list.append(
                [
                    os.path.join(self.save_raw_h5, imagename),
                    os.path.join(self.save_real_mask_h5, imagename),
                ]
            )
        create_csv(
            data_list,
            self.base_dir,
            save_train=self.save_train,
            save_test=self.save_test,
            save_val=self.save_val,
        )

    def _train_h5(self):

        self.train_list = os.path.join(self.base_dir, self.save_train)
        self.val_list = os.path.join(self.base_dir, self.save_val)
        self.test_list = os.path.join(self.base_dir, self.save_test)

        print(len(self.train_list), len(self.val_list), len(self.test_list))
        hparams = {
            "train_list": self.train_list,
            "test_list": self.test_list,
            "val_list": self.val_list,
            "data_root": "",
            "patch_size": self.patch_size,
            "epochs": self.epochs,
            "image_groups": ("data/image",),
            "mask_groups": (
                "data/distance",
                "data/flow_x",
                "data/flow_y",
                "data/flow_z",
            ),
            "dist_handling": "bool_inv",
            "in_channels": self.in_channels,
            "out_channels": self.out_channels,
            "feat_channels": tuple(self.feat_channels),
            "num_levels": self.num_levels,
            "norm_method": "instance",
            "samples_per_epoch": self.samples_per_epoch,
            "batch_size": self.batch_size,
            "learning_rate": self.learning_rate,
            "background_weight": self.background_weight,
            "flow_weight": self.flow_weight,
            "out_activation": self.out_activation,
        }

        save_json(
            hparams, str(self.base_dir) + "/" + self.model_name + ".json"
        )

        train_dataset = TrainTiled(
            hparams["train_list"],
            hparams["data_root"],
            patch_size=hparams["patch_size"],
            image_groups=hparams["image_groups"],
            mask_groups=hparams["mask_groups"],
            dist_handling=hparams["dist_handling"],
            samples_per_epoch=hparams["samples_per_epoch"],
        )
        train_loader = DataLoader(
            train_dataset,
            batch_size=hparams["batch_size"],
            shuffle=False,
            drop_last=True,
        )

        val_dataset = TrainTiled(
            hparams["val_list"],
            hparams["data_root"],
            patch_size=hparams["patch_size"],
            image_groups=hparams["image_groups"],
            mask_groups=hparams["mask_groups"],
            dist_handling=hparams["dist_handling"],
            samples_per_epoch=hparams["samples_per_epoch"],
        )
        val_loader = DataLoader(
            val_dataset,
            batch_size=hparams["batch_size"],
            shuffle=False,
            drop_last=True,
        )
        self.save_raw_h5 = os.path.join(self.base_dir, self.save_raw_h5_name)
        Path(self.save_raw_h5).mkdir(exist_ok=True)

        self.save_real_mask_h5 = os.path.join(
            self.base_dir, self.save_real_mask_h5_name
        )
        Path(self.save_real_mask_h5).mkdir(exist_ok=True)

        if self.network_type == "unet":
            self.model = CellPose3DModel(hparams)
        else:
            self.model = CellPoseRes3DModel(hparams)

        checkpoint_callback = ModelCheckpoint(
            dirpath=Path(self.model_dir),
            filename=self.model_name,
            save_top_k=1,
            monitor="epoch",
            mode="max",
            verbose=True,
        )

        logger = CSVLogger(
            save_dir=self.base_dir,
            name="lightning_logs_" + self.model_name,
        )

        trainer = Trainer(
            accelerator="auto",
            devices="auto",
            strategy="auto",
            logger=logger,
            callbacks=checkpoint_callback,
            min_epochs=self.epochs,
            max_epochs=self.epochs * 2,
            default_root_dir=self.model_dir,
            num_sanity_val_steps=0,
            enable_checkpointing=True,
        )

        if self.ckpt_path is not None:
            trainer.fit(
                self.model,
                train_dataloaders=train_loader,
                val_dataloaders=val_loader,
                ckpt_path=self.ckpt_path,
            )
        else:
            trainer.fit(
                self.model,
                train_dataloaders=train_loader,
                val_dataloaders=val_loader,
            )



import numpy as np
import os

# from IPython.display import clear_output
from scipy.ndimage import binary_fill_holes
from scipy.ndimage.measurements import find_objects
from scipy.ndimage import binary_dilation
from csbdeep.io import load_training_data
from csbdeep.utils import axes_dict
from csbdeep.models import ProjectionConfig, ProjectionCARE
from tifffile import imread
from csbdeep.data import RawData, create_patches_reduced_target
from skimage.measure import regionprops
from scipy import ndimage


def _raise(e):
    raise e


def _fill_label_holes(lbl_img, **kwargs):
    lbl_img_filled = np.zeros_like(lbl_img)
    for lb in set(np.unique(lbl_img)) - {0}:
        mask = lbl_img == lb
        mask_filled = binary_fill_holes(mask, **kwargs)
        lbl_img_filled[mask_filled] = lb
    return lbl_img_filled


def fill_label_holes(lbl_img, **kwargs):
    """Fill small holes in label image."""
    # TODO: refactor 'fill_label_holes' and 'edt_prob' to share code
    def grow(sl, interior):
        return tuple(
            slice(s.start - int(w[0]), s.stop + int(w[1])) for s, w in zip(sl, interior)
        )

    def shrink(interior):
        return tuple(slice(int(w[0]), (-1 if w[1] else None)) for w in interior)

    objects = find_objects(lbl_img)
    lbl_img_filled = np.zeros_like(lbl_img)
    for i, sl in enumerate(objects, 1):
        if sl is None:
            continue
        interior = [(s.start > 0, s.stop < sz) for s, sz in zip(sl, lbl_img.shape)]
        shrink_slice = shrink(interior)
        grown_mask = lbl_img[grow(sl, interior)] == i
        mask_filled = binary_fill_holes(grown_mask, **kwargs)[shrink_slice]
        lbl_img_filled[sl][mask_filled] = i
    return lbl_img_filled


def dilate_label_holes(lbl_img, iterations):
    lbl_img_filled = np.zeros_like(lbl_img)
    for lb in range(np.min(lbl_img), np.max(lbl_img) + 1):
        mask = lbl_img == lb
        mask_filled = binary_dilation(mask, iterations=iterations)
        lbl_img_filled[mask_filled] = lb
    return lbl_img_filled


def erode_labels(segmentation, erosion_iterations=2):
    # create empty list where the eroded masks can be saved to
    regions = regionprops(segmentation)
    erode = np.zeros(segmentation.shape)

    def erode_mask(segmentation_labels, label_id, erosion_iterations):

        only_current_label_id = np.where(segmentation_labels == label_id, 1, 0)
        eroded = ndimage.binary_erosion(
            only_current_label_id, iterations=erosion_iterations
        )
        relabeled_eroded = np.where(eroded == 1, label_id, 0)
        return relabeled_eroded

    for i in range(len(regions)):
        label_id = regions[i].label
        erode = erode + erode_mask(segmentation, label_id, erosion_iterations)

    # convert list of numpy arrays to stacked numpy array
    return erode


class Projection3D:
    def __init__(
        self,
        base_dir,
        npz_filename,
        model_name,
        model_dir,
        n_patches_per_image,
        raw_dir="Raw/",
        projection_dir="projection/",
        val_raw_dir="val_raw/",
        val_projection_dir="val_projection/",
        n_channel_in=1,
        downsample_factor=1,
        generate_npz=True,
        validation_split=0.01,
        patch_x=256,
        patch_y=256,
        grid_x=1,
        grid_y=1,
        use_gpu=True,
        batch_size=4,
        depth=3,
        kern_size=3,
        startfilter=48,
        epochs=400,
        learning_rate=0.0001,
    ):

        self.npz_filename = npz_filename
        self.base_dir = base_dir
        self.downsample_factor = downsample_factor
        self.model_dir = model_dir
        self.raw_dir = raw_dir
        self.projection_dir = projection_dir
        self.val_raw_dir = val_raw_dir
        self.val_projection_dir = val_projection_dir
        self.projection_dir = projection_dir
        self.generate_npz = generate_npz
        self.model_name = model_name
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.depth = depth
        self.n_channel_in = n_channel_in
        self.kern_size = kern_size
        self.patch_x = patch_x
        self.patch_y = patch_y
        self.grid_x = grid_x
        self.grid_y = grid_y
        self.validation_split = validation_split
        self.batch_size = batch_size
        self.use_gpu = use_gpu
        self.startfilter = startfilter
        self.n_patches_per_image = n_patches_per_image

        self.Train()

    def Train(self):

        if self.generate_npz:

            raw_data = RawData.from_folder(
                basepath=self.base_dir,
                source_dirs=[self.raw_dir],
                target_dir=self.projection_dir,
                axes="ZYX",
            )

            X, Y, XY_axes = create_patches_reduced_target(
                raw_data=raw_data,
                patch_size=(None, self.patch_y, self.patch_x),
                n_patches_per_image=self.n_patches_per_image,
                target_axes="YX",
                reduction_axes="Z",
                save_file=self.base_dir + self.npz_filename + ".npz",
            )

        print("Training UNET model")
        load_path = self.base_dir + self.npz_filename + ".npz"

        (X, Y), (X_val, Y_val), axes = load_training_data(
            load_path, validation_split=self.validation_split, verbose=True
        )
        c = axes_dict(axes)["C"]
        n_channel_in, n_channel_out = X.shape[c], Y.shape[c]

        config = ProjectionConfig(
            axes,
            n_channel_in,
            n_channel_out,
            unet_n_depth=self.depth,
            train_epochs=self.epochs,
            train_batch_size=self.batch_size,
            unet_n_first=self.startfilter,
            train_loss="mse",
            unet_kern_size=self.kern_size,
            train_learning_rate=self.learning_rate,
            train_reduce_lr={"patience": 5, "factor": 0.5},
        )
        print(config)
        vars(config)

        model = ProjectionCARE(
            config, name="projection_" + self.model_name, basedir=self.model_dir
        )

        if os.path.exists(
            self.model_dir + "projection_" + self.model_name + "/" + "weights_now.h5"
        ):
            print("Loading checkpoint model")
            model.load_weights(
                self.model_dir
                + "projection_"
                + self.model_name
                + "/"
                + "weights_now.h5"
            )

        if os.path.exists(
            self.model_dir + "projection_" + self.model_name + "/" + "weights_last.h5"
        ):
            print("Loading checkpoint model")
            model.load_weights(
                self.model_dir
                + "projection_"
                + self.model_name
                + "/"
                + "weights_last.h5"
            )

        if os.path.exists(
            self.model_dir + "projection_" + self.model_name + "/" + "weights_best.h5"
        ):
            print("Loading checkpoint model")
            model.load_weights(
                self.model_dir
                + "projection_"
                + self.model_name
                + "/"
                + "weights_best.h5"
            )

        history = model.train(X, Y, validation_data=(X_val, Y_val))

        print(sorted(list(history.history.keys())))


def read_float(fname):

    return imread(fname).astype("float32")



from setuptools import find_packages, setup
from os import path

try:
    from ... import version as __version__
except ImportError:
    __version__ = "unknown"
_dir = path.dirname(__file__)


with open("README.md") as f:
    long_description = f.read()
with open(
    path.join(_dir, "src/vollseg", "_version.py"), encoding="utf-8"
) as f:
    exec(f.read())

setup(
    name="vollseg",
    version=__version__,
    author="Varun Kapoor,Claudia Carabana Garcia,Mari Tolonen, Jakub Sedzinski",
    author_email="randomaccessiblekapoor@gmail.com",
    url="https://github.com/Kapoorlabs-CAPED/VollSeg/",
    description="Segmentation tool for biological cells of irregular size and shape in 3D and 2D, using \
    StarDist, U-NET, CARE, CellPose and SAM",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "pandas",
        "stardist",
        "scipy",
        "tifffile",
        "matplotlib",
        "napari",
        "cellpose-vollseg",
        "torch",
        "torchvision",
        "torchaudio",
        "test_tube",
        "lightning",
        "segment-anything",
        "opencv-contrib-python-headless",
        "cellpose",
    ],
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
    ],
)



from csbdeep.models import CARE
from csbdeep.internals import train
from .pretrained import (
    get_registered_models,
    get_model_details,
    get_model_instance,
)
import sys

# if IS_TF_1:
#     import tensorflow as tf
# else:
#     import tensorflow.compat.v1 as tf
#     # tf.disable_v2_behavior()


class CARE(CARE):
    """Standard CARE network for image restoration and enhancement.

    Uses a convolutional neural network created by :func:`csbdeep.internals.nets.common_unet`.
    Note that isotropic reconstruction and manifold extraction/projection are not supported here
    (see :class:`csbdeep.models.IsotropicCARE` ).

    Parameters
    ----------
    config : :class:`csbdeep.models.Config` or None
        Valid configuration of CARE network (see :func:`Config.is_valid`).
        Will be saved to disk as JSON (``config.json``).
        If set to ``None``, will be loaded from disk (must exist).
    name : str or None
        Model name. Uses a timestamp if set to ``None`` (default).
    basedir : str
        Directory that contains (or will contain) a folder with the given model name.
        Use ``None`` to disable saving (or loading) any data to (or from) disk (regardless of other parameters).

    Raises
    ------
    FileNotFoundError
        If ``config=None`` and config cannot be loaded from disk.
    ValueError
        Illegal arguments, including invalid configuration.

    Example
    -------
    >>> model = CARE(config, 'my_model')

    Attributes
    ----------
    config : :class:`csbdeep.models.Config`
        Configuration of CARE network, as provided during instantiation.
    keras_model : `Keras model <https://keras.io/getting-started/functional-api-guide/>`_
        Keras neural network model.
    name : str
        Model name.
    logdir : :class:`pathlib.Path`
        Path to model folder (which stores configuration, weights, etc.)
    """

    def __init__(self, config, name=None, basedir="."):
        """See class docstring."""
        super().__init__(config=config, name=name, basedir=basedir)

    @classmethod
    def local_from_pretrained(cls, name_or_alias=None):
        try:
            get_model_details(cls, name_or_alias, verbose=True)
            return get_model_instance(cls, name_or_alias)
        except ValueError:
            if name_or_alias is not None:
                print(
                    "Could not find model with name or alias '%s'"
                    % (name_or_alias),
                    file=sys.stderr,
                )
                sys.stderr.flush()
            get_registered_models(cls, verbose=True)

    def train(
        self,
        X,
        Y=None,
        validation_data=None,
        steps_per_epoch=None,
        load_data_sequence=False,
    ):

        epochs = self.config.train_epochs

        if not self._model_prepared:
            self.prepare_for_training()

        if not load_data_sequence:
            # small training data
            steps_per_epoch = len(X) // self.config.train_batch_size
            training_data = train.DataWrapper(
                X,
                Y,
                self.config.train_batch_size,
                length=epochs * steps_per_epoch,
            )

            history = self.keras_model.fit(
                iter(training_data),
                validation_data=validation_data,
                epochs=epochs,
                callbacks=self.callbacks,
                verbose=1,
            )
        else:
            # A sequence object
            history = self.keras_model.fit(
                X,
                validation_data=validation_data,
                epochs=epochs,
                callbacks=self.callbacks,
                verbose=1,
            )
        self._training_finished()

        return history
